@@ -21,15 +21,19 @@ LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 SOFTWARE.
 """
+from datetime import datetime
+from os import path
+import time
+from typing import Union
 
 import cv2
 import numpy as np
 import tensorflow as tf
-import time
-from typing import Union
+from tensorflow.keras import backend, layers, models
 
-from ..utility import dataset, media, predict, train, utils, weights
+from . import dataset, train, weights
 from ..model import yolov4
+from ..utility import media, predict
 
 
 class YOLOv4:
@@ -38,25 +42,12 @@ class YOLOv4:
         Default configuration
         """
         self.anchors = [
-            12,
-            16,
-            19,
-            36,
-            40,
-            28,
-            36,
-            75,
-            76,
-            55,
-            72,
-            146,
-            142,
-            110,
-            192,
-            243,
-            459,
-            401,
+            [[12, 16], [19, 36], [40, 28]],
+            [[36, 75], [76, 55], [72, 146]],
+            [[142, 110], [192, 243], [459, 401]],
         ]
+        self.batch_size = 32
+        self.subdivision = 16
         self._classes = None
         self._has_weights = False
         self.input_size = 608
@@ -98,7 +89,7 @@ class YOLOv4:
     @classes.setter
     def classes(self, data: Union[str, dict]):
         if isinstance(data, str):
-            self._classes = utils.read_class_names(data)
+            self._classes = media.read_classes_names(data)
         elif isinstance(data, dict):
             self._classes = data
         else:
@@ -155,45 +146,69 @@ class YOLOv4:
             self._xyscales = xyscales
 
     def make_model(self):
+        # pylint: disable=missing-function-docstring
         self._has_weights = False
-        tf.keras.backend.clear_session()
+        backend.clear_session()
+
+        inputs = layers.Input([self.input_size, self.input_size, 3])
         self.model = yolov4.YOLOv4(
             anchors=self.anchors,
-            input_size=self.input_size,
             num_classes=len(self.classes),
             xyscales=self.xyscales,
         )
-        # [batch, height, width, channel]
-        self.model(tf.keras.layers.Input([self.input_size, self.input_size, 3]))
+        self.model(inputs)
 
-    def load_weights(self, path: str, weights_type: str = "tf"):
+    def load_weights(self, weights_path: str, weights_type: str = "tf"):
         """
         Usage:
             yolo.load_weights("yolov4.weights", weights_type="yolo")
             yolo.load_weights("checkpoints")
         """
         if weights_type == "yolo":
-            weights.load_yolov4(self.model, path)
+            weights.load_weights(self.model, weights_path)
         elif weights_type == "tf":
-            self.model.load_weights(path).expect_partial()
+            self.model.load_weights(weights_path).expect_partial()
 
         self._has_weights = True
 
+    def save_as_tflite(self, tflite_path):
+        """
+        Save model and weights as tflite
+
+        Usage:
+            yolo.save_as_tflite("yolov4.tflite")
+        """
+        converter = tf.lite.TFLiteConverter.from_keras_model(self.model)
+        converter.allow_custom_ops = True
+        tflite_model = converter.convert()
+        with tf.io.gfile.GFile(tflite_path, "wb") as fd:
+            fd.write(tflite_model)
+
     def predict(self, frame: np.ndarray):
         image_data = media.resize(frame, self.input_size)
         image_data = image_data / 255
         image_data = image_data[np.newaxis, ...].astype(np.float32)
 
         candidates = self.model.predict(image_data)
-        candidates = predict.reduce_bbox_candidates(
-            candidates[0], self.input_size
-        )
+        _candidates = []
+        for candidate in candidates:
+            batch_size = candidate.shape[0]
+            grid_size = candidate.shape[1]
+            _candidates.append(
+                tf.reshape(
+                    candidate, shape=(batch_size, grid_size * grid_size * 3, -1)
+                )
+            )
+        candidates = np.concatenate(_candidates, axis=1)
+        candidates = predict.reduce_bbox_candidates(candidates, self.input_size)
         candidates = predict.fit_predicted_bboxes_to_original(
             candidates, frame.shape
         )
         return candidates
 
     def inference(self, media_path, is_image=True, cv_waitKey_delay=10):
+        if not path.exists(media_path):
+            raise FileNotFoundError("{} does not exist".format(media_path))
         if is_image:
             frame = cv2.imread(media_path)
             frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
@@ -206,7 +221,7 @@ class YOLOv4:
             print(info)
 
             frame = cv2.cvtColor(frame, cv2.COLOR_RGB2BGR)
-            image = media.draw_bbox(frame, bboxes, self.classes)
+            image = media.draw_bbox(frame, bboxes[0], self.classes)
             cv2.namedWindow("result", cv2.WINDOW_AUTOSIZE)
             cv2.imshow("result", image)
         else:
@@ -226,7 +241,7 @@ class YOLOv4:
                 print(info)
 
                 frame = cv2.cvtColor(frame, cv2.COLOR_RGB2BGR)
-                image = media.draw_bbox(frame, bboxes, self.classes)
+                image = media.draw_bbox(frame, bboxes[0], self.classes)
                 cv2.namedWindow("result", cv2.WINDOW_AUTOSIZE)
                 cv2.imshow("result", image)
                 if cv2.waitKey(cv_waitKey_delay) & 0xFF == ord("q"):
@@ -237,35 +252,41 @@ class YOLOv4:
             pass
         cv2.destroyWindow("result")
 
-    def load_datasets(
-        self, datasets_path, datasets_type="converted_coco", training=True
+    def load_dataset(
+        self, dataset_path, dataset_type="converted_coco", training=True
     ):
         return dataset.Dataset(
             anchors=self.anchors,
-            datasets_path=datasets_path,
-            datasets_type=datasets_type,
+            batch_size=self.batch_size // self.subdivision,
+            dataset_path=dataset_path,
+            dataset_type=dataset_type,
+            data_augmentation=training,
             input_size=self.input_size,
             num_classes=len(self.classes),
             strides=self.strides,
-            training=training,
+            xyscales=self.xyscales,
         )
 
-    def compile(self, iou_type: str = "giou", learning_rate: float = 1e-5):
-        self.model.compile(iou_type=iou_type, learning_rate=learning_rate)
+    def compile(self, iou_type: str = "ciou", learning_rate: float = 1e-4):
+        self.model.compile(
+            optimizer=tf.keras.optimizers.Adam(learning_rate=learning_rate),
+            loss=train.YOLOv4Loss(
+                batch_size=self.batch_size // self.subdivision,
+                iou_type=iou_type,
+            ),
+        )
 
-    def fit(self, datasets, epochs, batch_size: int = 6):
-        for epoch in range(epochs):
-            batch = []
-            for i in range(batch_size):
-                batch.append(next(datasets))
-
-            batchset = (
-                tf.concat([x[0] for x in batch], axis=0),
-                tf.concat([x[1] for x in batch], axis=0),
-            )
-            loss = self.model.train_step(batchset)
-            print(
-                "epoch: {: 4}, batch: {: 4}, _iou_loss: {:7.2f}, score_loss: {:7.2f}, classes_loss: {:7.2f}, loss: {:7.2f}".format(
-                    epoch, i + 1, loss[0], loss[1], loss[2], loss[3]
-                )
+    def fit(self, data_set, epochs, log_dir_path=None):
+        callbacks = []
+        if log_dir_path:
+            tensorboard_callback = tf.keras.callbacks.TensorBoard(
+                log_dir=log_dir_path, histogram_freq=1
             )
+            callbacks.append(tensorboard_callback)
+        self.model.fit(
+            data_set,
+            epochs=epochs,
+            batch_size=data_set.batch_size,
+            steps_per_epoch=self.subdivision,
+            callbacks=callbacks,
+        )
